@@ -1,7 +1,7 @@
 /*
    Various utilities
 
-   Copyright (C) 1994-2024
+   Copyright (C) 1994-2025
    Free Software Foundation, Inc.
 
    Written by:
@@ -459,9 +459,19 @@ size_trunc_len (char *buffer, unsigned int len, uintmax_t size, int units, gbool
         { "", "k", "m", "g", "t", "p", "e", "z", "y", "r", "q", NULL };
     /* *INDENT-ON* */
 
+    static int sfx_last = -1;
+
     const char *const *sfx = use_si ? suffix_lc : suffix;
     int j = 0;
 
+    if (sfx_last < 0)
+    {
+        for (sfx_last = 0; sfx[sfx_last] != NULL; sfx_last++)
+            ;
+
+        sfx_last--;
+    }
+
     if (len == 0)
         len = 9;
 #if SIZEOF_UINTMAX_T == 8
@@ -474,13 +484,15 @@ size_trunc_len (char *buffer, unsigned int len, uintmax_t size, int units, gbool
         len = 9;
 #endif
 
+    const int units_safe = MIN (units, sfx_last);
+
     /*
      * recalculate from 1024 base to 1000 base if units>0
      * We can't just multiply by 1024 - that might cause overflow
      * if uintmax_t type is too small
      */
     if (use_si)
-        for (j = 0; j < units; j++)
+        for (j = 0; j < units_safe; j++)
         {
             uintmax_t size_remain;
 
@@ -490,7 +502,7 @@ size_trunc_len (char *buffer, unsigned int len, uintmax_t size, int units, gbool
             size += size_remain;        /* Re-add remainder lost by division/multiplication */
         }
 
-    for (j = units; sfx[j] != NULL; j++)
+    for (j = units_safe; sfx[j] != NULL; j++)
     {
         if (size == 0)
         {
@@ -635,7 +647,7 @@ load_mc_home_file (const char *from, const char *filename, char **allocated_file
 /* --------------------------------------------------------------------------------------------- */
 
 const char *
-extract_line (const char *s, const char *top)
+extract_line (const char *s, const char *top, size_t *len)
 {
     static char tmp_line[BUF_MEDIUM];
     char *t = tmp_line;
@@ -643,6 +655,10 @@ extract_line (const char *s, const char *top)
     while (*s != '\0' && *s != '\n' && (size_t) (t - tmp_line) < sizeof (tmp_line) - 1 && s < top)
         *t++ = *s++;
     *t = '\0';
+
+    if (len != NULL)
+        *len = (size_t) (t - tmp_line);
+
     return tmp_line;
 }
 
@@ -732,7 +748,10 @@ skip_numbers (const char *s)
  * "control sequence", in a sort of pidgin BNF, as follows:
  *
  * control-seq = Esc non-'['
- *             | Esc '[' (0 or more digits or ';' or ':' or '?') (any other char)
+ *             | Esc '[' (parameter-byte)* (intermediate-byte)* final-byte
+ * parameter-byte = [\x30-\x3F]     # one of "0-9;:<=>?"
+ * intermediate-byte = [\x20–\x2F]  # one of " !\"#$%&'()*+,-./"
+ * final-byte = [\x40-\x7e]         # one of "@A–Z[\]^_`a–z{|}~"
  *
  * The 256-color and true-color escape sequences should allow either ';' or ':' inside as separator,
  * actually, ':' is the more correct according to ECMA-48.
@@ -759,8 +778,10 @@ strip_ctrl_codes (char *s)
             if (*(++r) == '[' || *r == '(')
             {
                 /* strchr() matches trailing binary 0 */
-                while (*(++r) != '\0' && strchr ("0123456789;:?", *r) != NULL)
+                while (*(++r) != '\0' && strchr ("0123456789;:<=>?", *r) != NULL)
                     ;
+                while (*r != '\0' && (*r < 0x40 || *r > 0x7E))
+                    ++r;
             }
             else if (*r == ']')
             {
